/*
 * Copyright (C) 2023 PulkoMandy <pulkomandy@pullkomandy.tk>
 *
 * Distributed under terms of the MIT license.
 */

#include "HyperTextView.h"

#include <Application.h>
#include <Button.h>
#include <ByteOrder.h>
#include <Directory.h>
#include <Entry.h>
#include <File.h>
#include <GroupLayout.h>
#include <LayoutBuilder.h>
#include <String.h>
#include <TextEncoding.h>
#include <TextView.h>
#include <Window.h>

#include <errno.h>
#include <stdio.h>

#include <string>
#include <vector>

#define BOLD "\x1B[34m"
#define NORMAL "\x1B[0m"

void disassemble(const std::vector<uint8_t>& script)
{
	size_t pc = 0;

	while (pc < script.size()) {
		uint8_t op = script[pc];
		switch(op) {
			case 0x00:
				printf(BOLD "Return ");
				pc += 1;
				break;
			case 0x01:
				printf(BOLD "CheckFlag(%02x): ", script[pc + 1]);
				pc += 2;
				break;
			case 0x03:
				printf(BOLD "HaveItem(%02x): ", script[pc + 1]);
				pc += 2;
				break;
			case 0x09:
				printf(BOLD "IfEqual(%02x, %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				pc += 4;
				break;
			case 0x0A:
				printf(BOLD "IfLess(%02x, %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				pc += 4;
				break;
			case 0x0D:
				printf(BOLD "Hook(%02x%02x): ", script[pc + 2], script[pc + 1]);
				pc += 3;
				break;
			case 0x0E:
				printf(BOLD "GlobalHook(%02x): ", script[pc + 1]);
				pc += 2;
				break;
			case 0x7E:
				printf(BOLD "Or ");
				pc += 1;
				break;
			case 0x7F:
				printf(BOLD "Not "); // is it And?
				pc += 1;
				break;
			case 0x80:
				printf(NORMAL "80(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x81:
				printf(BOLD "SetFlag(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x82:
				printf(BOLD "ClearFlag(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x83:
				printf(BOLD "GetItem(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x85:
				printf(BOLD "LoseItem(%02x %02x) ", script[pc + 1], script[pc + 2]);
				pc += 3;
				break;
			case 0x87:
				printf(BOLD "SetCounter(%02x, %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				pc += 4;
				break;
			case 0x89:
				printf(BOLD "Add(%02x, %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				pc += 4;
				break;
			case 0x8A:
				printf(BOLD "Sub(%02x, %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				pc += 4;
				break;
			case 0x8F:
				printf(BOLD "GoRoom(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x90:
				printf(BOLD "GoScreen(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x91:
				printf(BOLD "ShowGlobalScreen(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x93:
				// Together with a GoRoom, allows to enter a room in a specific screen
				printf(BOLD "WithScreen(%02x) ", script[pc + 1]);
				pc += 2;
				break;
			case 0x95:
				printf(BOLD "InitFlag(%02x, %02x) ", script[pc + 1], script[pc + 2]);
				pc += 3;
				break;
			case 0xA0:
				printf(BOLD "Compare(%02x, %02x%02x) ", script[pc + 1], script[pc + 3], script[pc + 2]);
				pc += 4;
				break;
			case 0xFA:
				printf(BOLD "End.");
				pc++;
				break;
			case 0xFC:
				printf(BOLD "Game Over ");
				pc++;
				break;
			case 0xFD:
				printf(BOLD "Next Script ");
				pc++;
				break;
			case 0xFE:
				printf(BOLD "Goto(%02x)", script[pc + 1]);
				pc += 2;
				break;
			case 0xFF:
				printf(BOLD "You won! ");
				pc++;
				break;
			default:
				printf(NORMAL "%02x ", op);
				pc++;
				break;
		}
	}
	puts(NORMAL);
}

const char* cp850_lookup[0x80] = {
	"Ç", "ü", "é", "â", "ä", "à", "å", "ç", "ê", "ë", "è", "ï", "î", "ì", "Ä", "Å", "É", "æ", "Æ",
	"ô", "ö", "ò", "û", "ù", "ÿ", "Ö", "Ü", "ø", "£", "Ø", "×", "ƒ", "á", "í", "ó", "ú", "ñ", "Ñ",
	"ª", "º", "¿", "®", "¬", "½", "¼", "¡", "«", "»", "░", "▒", "▓", "│", "┤", "Á", "Â", "À", "©",
	"╣", "║", "╗", "╝", "¢", "¥", "┐", "└", "┴", "┬", "├", "─", "┼", "ã", "Ã", "╚", "╔", "╩", "╦",
	"╠", "═", "╬", "¤", "ð", "Ð", "Ê", "Ë", "È", "ı", "Í", "Î", "Ï", "┘", "┌", "█", "▄", "¦", "Ì",
	"▀", "Ó", "ß", "Ô", "Ò", "õ", "Õ", "µ", "þ", "Þ", "Ú", "Û", "Ù", "ý", "Ý", "¯", "´", "\u00AD",
	"±", "‗", "¾", "¶", "§", "÷", "¸", "°", "¨", "·", "¹", "³", "²", "■", "\u00A0"
};

class Room {
	public:
	std::vector<std::string> fMessages;
	std::vector<std::vector<uint8_t>> fScripts;
};

std::vector<Room> gRooms;
std::vector<uint8_t> gVar8;
std::vector<uint16_t> gVar16;

uint8_t gCurrentRoom;
uint8_t gCurrentScreen;
uint8_t gTrigger;

// TODO objects
// TODO actions
// TODO variables (c and f) and their init
// TODO global buttons
// TODO initial room
// TODO other things from the init script

class GameWindow: public BWindow
{
	public:
		GameWindow()
			: BWindow(BRect(40, 40, 500, 400), "Readingame", B_DOCUMENT_WINDOW,
				B_AUTO_UPDATE_SIZE_LIMITS | B_QUIT_ON_WINDOW_CLOSE)
			, fDecoder("cp-850")
		{
			fMainText = new HyperTextView("main text", B_WILL_DRAW);

			// TODO making it non-editable while having custom insets breaks the layout somehow?
			fMainText->MakeEditable(false);
			fMainText->MakeSelectable(false);
			//fMainText->SetInsets(10, 10, 10, 10);

			fInlineButton = new BButton("button", NULL);

			BLayoutBuilder::Group<>(this, B_VERTICAL)
				.SetInsets(0, 0, 0, B_USE_WINDOW_SPACING)
				.Add(fMainText, 999)
				.AddGroup(B_HORIZONTAL)
					.AddGlue()
					.Add(fInlineButton)
					.AddGlue()
				.End()
				.AddGlue(0)
			.End();
		}

		void Show() override
		{
			BWindow::Show();
			Lock();
			fInlineButton->SetLowColor(make_color(255, 128, 255));
			fMainText->SetViewColor(make_color(196, 255, 255));
			fInlineButton->SetTarget(be_app);
			Unlock();
		}

		void SetMainText(const std::string& text)
		{
			Lock();
			fInlineButton->Hide();
			fMainText->SetText("");

			BString decoded;
			PrettyPrint(text, decoded);

			text_run_array runs;
			runs.count = 1;
			runs.runs->font = be_plain_font;
			runs.runs->offset = 0;
			runs.runs->color = make_color(0, 0, 0);
			fMainText->Insert(decoded, &runs);
			fMainText->Invalidate();
			Unlock();
		}

		void PrettyPrint(const std::string& message, BString& decoded)
		{
			size_t cursor = 0;
			while (cursor < message.size()) {
				switch(message[cursor]) {
					case '\\':
						cursor += HandleEscape(message, cursor + 1);
						break;
					case '[':
					{
						text_run_array runs;
						runs.count = 1;
						runs.runs->font = be_plain_font;
						runs.runs->offset = 0;
						runs.runs->color = make_color(0, 0, 0);
						fMainText->Insert(decoded, &runs);
						decoded = "";
						cursor += HandleHyperlink(message, cursor);
						break;
					}
					case '@':
					case '$':
						// @ Gender mark (print an e if the player identifies as female)
						// $ Plural mark (print an s if the counter is plural)
						printf(BOLD "%c" NORMAL, message[cursor]);
						cursor++;
						break;
					case '_':
						// Used as a non-breakable space
						decoded += "\u00A0";
						cursor++;
						break;
					default:
						if (message[cursor] >= 0)
							decoded += message[cursor];
						else
							decoded += cp850_lookup[(message[cursor] - 0x80) & 0xFF];
						cursor++;
						break;
				}
			}
			decoded += '\n';
		}

		int HandleHyperlink(const std::string& message, int cursor)
		{
			BString linkText;

			int v = 0;
			int i = 1;
			while (isdigit(message[cursor + i])) {
				v = v * 10 + message[cursor + i] - '0';
				i++;
			}

			while (message[cursor + i] != ']') {
				if (message[cursor + i] >= 0)
					linkText += message[cursor + i];
				else
					linkText += cp850_lookup[(message[cursor + i] - 0x80) & 0xFF];
				i++;
			}

			text_run_array single_run;
			single_run.count = 1;
			single_run.runs->offset = 0;
			single_run.runs->font = be_plain_font;
			single_run.runs->color = make_color(0, 0, 196);
			fMainText->InsertHyperText(linkText, new HyperTextAction(v), &single_run);

			return i + 1;
		}

		int HandleEscape(const std::string& message, int cursor)
		{
			if (message[cursor] == 'o') {
				fInlineButton->SetLabel("OK");
				fInlineButton->Show();
				int v = 0;
				int i = 1;
				while (isdigit(message[cursor + i])) {
					v = v * 10 + message[cursor + i] - '0';
					i++;
				}
				fInlineButton->SetMessage(new BMessage(v));
				return i + 1;
			} else if (message[cursor] == '!') {
				printf(BOLD "APPEND" NORMAL "\n");
				return 2;
			} else {
				// TODO unknown escape sequence
				fprintf(stderr, "ERROR: unknown escape sequence %c", message[cursor]);
				return 2;
			}
		}

	private:
		HyperTextView* fMainText;
		BButton* fInlineButton;
		BPrivate::BTextEncoding fDecoder;
};

class Readingame: public BApplication
{
	public:
	Readingame()
		: BApplication("application/x-vnd.PulkoMandy.Readingame")
	{
	}

	void ReadyToRun() override
	{
		// TODO check if we have a game loaded
		fMainWindow = new GameWindow();

		// FIXME add the global buttons to the main window

		BWindow* debugWindow = new BWindow(BRect(600, 40, 900, 200), "Readingame debugger",
			B_DOCUMENT_WINDOW, B_AUTO_UPDATE_SIZE_LIMITS | B_QUIT_ON_WINDOW_CLOSE);

		fMainWindow->Show();
		fMainWindow->CenterOnScreen();

		// TODO: only if asked from the command line
		debugWindow->Show();

		bool runNextLine = true;
		int line = 0;
		while (runNextLine) {
			runNextLine = RunScriptLine(gRooms[0].fScripts[line]);
			line++;
		}
		fMainWindow->SetMainText(gRooms[gCurrentRoom].fMessages[gCurrentScreen - 1]);
	}

	void MessageReceived(BMessage* message) override
	{
		if (message->what < 256) {
			gTrigger = message->what;
			int line = 0;
			bool runNextLine = true;
			while (runNextLine) {
				runNextLine = RunScriptLine(gRooms[0].fScripts[line]);
				line++;
			}
			line = 0;
			while (runNextLine) {
				runNextLine = RunScriptLine(gRooms[gCurrentRoom].fScripts[line]);
				line++;
			}
			fMainWindow->SetMainText(gRooms[gCurrentRoom].fMessages[gCurrentScreen - 1]);
		} else {
			BApplication::MessageReceived(message);
		}
	}

	bool RunScriptLine(const std::vector<uint8_t>& script)
	{
		int pc = 0;
		bool done = false;
		bool runNextLine = false;

		while (!done) {
			switch(script[pc++]) {
				case 0x00:
				{
					done = true;
					break;
				}
				case 0x09:
				{
					uint8_t varAddress = script[pc++];
					uint16_t varValue = script[pc++];
					varValue = varValue | (script[pc++] << 8);
					if (gVar16[varAddress] == varValue) {
						continue;
					} else {
						done = true;
						runNextLine = true;
					}
				}
				case 0x0D:
				{
					uint8_t eventToCheck = script[pc++];
					if (gTrigger == eventToCheck) {
						fprintf(stderr, "unknown param to 0D: %02x\n", script[pc++]);
						disassemble(script);
						continue;
					} else {
						done = true;
						runNextLine = true;
					}
					break;
				}
				case 0x0E:
				{
					uint8_t eventToCheck = script[pc++];
					// FIXME is it really gTrigger here, or should it be something else?
					if (gTrigger == eventToCheck)
						continue;
					else {
						done = true;
						runNextLine = true;
					}
					break;
				}
				case 0x7F:
				{
					// FIXME figure this out
					pc++;
					break;
				}
				case 0x87:
				{
					uint8_t varAddress = script[pc++];
					uint16_t varValue = script[pc++];
					varValue = varValue | (script[pc++] << 8);
					gVar16[varAddress] = varValue;
					break;
				}
				case 0x8F:
				{
					gCurrentRoom = script[pc++];
					break;
				}
				case 0x90:
				{
					gCurrentScreen = script[pc++];
					break;
				}
				case 0x93:
				{
					gCurrentScreen = script[pc++];
					break;
				}
				case 0x95:
				{
					uint8_t varAddress = script[pc++];
					uint8_t varValue = script[pc++];
					gVar8[varAddress] = varValue;
					break;
				}
				default:
					fprintf(stderr, "ERROR: unknown opcode %02x at %04x\n", script[pc - 1], pc - 1);
					disassemble(script);
					done = true;
					break;
			}
		}

		return runNextLine;
	}

private:
	GameWindow* fMainWindow;
};

void usage()
{
	fprintf(stderr,
		"An engine for playing game originally designed for the Lectures Enjeu application\n\n"
		"Usage: readingame path/to/gamedir/\n"
	);
}

void ReadRoomsScripts(BFile& f)
{
	static int roomNumber = 0;

	off_t fileSize;
	f.GetSize(&fileSize);
	while (fileSize > 0) {
		uint16_t roomSize;
		f.Read(&roomSize, 2);
		roomSize = B_LENDIAN_TO_HOST_INT16(roomSize);
		fileSize -= roomSize;

		if (roomSize == 0) {
			fileSize -= 2;
			break;
		}

		if (gRooms.size() < roomNumber + 1)
			gRooms.resize(roomNumber + 1);

		int i = 0;
		while (roomSize > 0) {
			uint16_t scriptSize;

			if (f.Read(&scriptSize, 2) != 2) {
				fprintf(stderr, "ERROR\n");
				return;
			}

			scriptSize = B_LENDIAN_TO_HOST_INT16(scriptSize);
			if (scriptSize == 0) {
				roomSize -= 2;
				break;
			}

			if (gRooms[roomNumber].fScripts.size() < i + 1)
				gRooms[roomNumber].fScripts.resize(i + 1);
			gRooms[roomNumber].fScripts[i].resize(scriptSize - 2);
			f.Read(gRooms[roomNumber].fScripts[i].data(), scriptSize - 2);

			roomSize -= scriptSize;
			i++;
		}
		roomNumber++;
	}
}

void ReadRoomsMessages(BFile& f)
{
	static int roomNumber = 0;

	off_t fileSize;
	f.GetSize(&fileSize);
	while (fileSize > 0) {
		uint16_t roomSize;
		f.Read(&roomSize, 2);
		roomSize = B_LENDIAN_TO_HOST_INT16(roomSize);
		fileSize -= roomSize;

		if (roomSize == 0) {
			fileSize -= 2;
			break;
		}

		roomSize -= 2;

		if (gRooms.size() < roomNumber + 1)
			gRooms.resize(roomNumber + 1);

		int i = 0;
		while (roomSize > 0) {
			uint16_t scriptSize;

			ssize_t r = f.Read(&scriptSize, 2);
			if (r != 2) {
				fprintf(stderr, "ERROR %d (%s)\n", r, strerror(errno));
				return;
			}
			scriptSize = B_LENDIAN_TO_HOST_INT16(scriptSize);
			if (scriptSize == 0) {
				roomSize -= 2;
				break;
			}

			if (gRooms[roomNumber].fMessages.size() < i + 1)
				gRooms[roomNumber].fMessages.resize(i + 1);
			gRooms[roomNumber].fMessages[i].resize(scriptSize - 2);
			f.Read(gRooms[roomNumber].fMessages[i].data(), scriptSize - 2);

			roomSize -= scriptSize;
			i++;
		}
		roomNumber++;
	}
}

int main(int argc, char** argv)
{
	if (argc != 2) {
		// TODO open a filepanel instead to load a game
		// TODO or just look for the TITRES.DAT file
		usage();
		return 1;
	}

	BDirectory gameDir(argv[1]);

	if (gameDir.InitCheck() != B_OK) {
		fprintf(stderr, "Game directory %s not found.\n", argv[1]);
		usage();
		return 1;
	}

	BEntry entry;
	BFile initFile;
	BFile msgRoomsFile;
	BFile msgGeneralFile;
	BFile condRoomsFile;
	BFile condGenFile;
	BFile condUltiFile;

	if (gameDir.FindEntry("INIT.DAT", &entry) == B_OK) {
		initFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain an INIT.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("MSG.DAT", &entry) == B_OK) {
		msgRoomsFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain an MSG.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("MSGGEN.DAT", &entry) == B_OK) {
		msgGeneralFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain an MSGGEN.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("CONDSAL.DAT", &entry) == B_OK) {
		condRoomsFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain a CONDSAL.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("CONDGALE.DAT", &entry) == B_OK) {
		condGenFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain a CONDGALE.DAT file.\n");
		return 2;
	}

	if (gameDir.FindEntry("CONDULTI.DAT", &entry) == B_OK) {
		condUltiFile.SetTo(&entry, B_READ_ONLY);
	} else {
		fprintf(stderr, "Game does not contain a CONDULTI.DAT file.\n");
		return 2;
	}

	ReadRoomsScripts(condGenFile);
	ReadRoomsScripts(condRoomsFile);
	ReadRoomsScripts(condUltiFile);

	ReadRoomsMessages(msgGeneralFile);
	ReadRoomsMessages(msgRoomsFile);

	// TODO how many variables are there?
	gVar8.resize(16);
	gVar16.resize(16);

	/*
	int i = 0;
	for (const auto& room: gRooms) {
		printf("========= ROOM %d =========\n", i);

		int j = 0;
		for (auto text: room.fMessages) {
			printf("------ SCREEN %d -------\n", j + 1);
			prettyPrint(text);
			j++;
		}

		j = 0;
		for (auto script: room.fScripts) {
			printf("%x:\t", j + 1);
			disassemble(script);
			j++;
		}
		i++;
	}
	*/

	Readingame app;

	app.Run();
}
